/*
   A simple electrostatic 3D Poisson solver for testing weak scaling.
   The problem is defined on a structured cubical grid with Dirichlet boundaries inside the domain.
   The grid's total size is determined by the number of processors and the number of nodes per dimension per process.
   Two arrays for the electric potential and the charge distribution are passed to an object of a PETSc solver class,
   from which KSPsolve() is called. The potential array is then updated and could principally be used to calculate
   the electric field.
*/

#include <iostream>
#include <mpi.h>
#include <stdlib.h>
#include <string.h>
#include <vector>

#include "petscsolver.h"


bool is_cube_number(int n){
  int croot = round(cbrt(n));
  return n == croot*croot*croot;
}

int iterations = 1000; //number of iterations of calls to solve()
int          N =   30; //number of nodes per dimension per process
double       h = 1E-4; //distance between grid points
 
int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);
  
  std::vector<std::string> all_args;
  if (argc > 1) all_args.assign(argv,argv+argc);
  
  int worldsize,rank;
  MPI_Comm_size(MPI_COMM_WORLD,&worldsize); //total number of processors
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  
  for (int i=1; i<argc-1; i++) {
    if (all_args[i]=="-nodes_per_proc")
      N = std::stoi(all_args[i+1]);
    if (all_args[i]=="-mesh_size")
      h = std::stod(all_args[i+1]);
    if (all_args[i]=="-iterations")
      iterations = std::stoi(all_args[i+1]);
  }
  
  /* determine largest cube number <= worldsize */
  int petscsize = worldsize;
  while (!is_cube_number(petscsize))
    petscsize--;
    
  int procs_per_dim = round(cbrt(petscsize)); // cubical process grid
  int g_N = procs_per_dim*N; //number of nodes per dimension
  int dims[3] = {procs_per_dim,procs_per_dim,procs_per_dim};
  
  MPI_Comm comm;
  int petsc_involved = 0;
  if (rank<petscsize) petsc_involved = 1;
  int key = rank;
  MPI_Comm_split(MPI_COMM_WORLD,petsc_involved,key,&comm);
  PETSC_COMM_WORLD = comm;
  
  
  PetscErrorCode ierr;
  ierr=PetscInitialize(&argc,&argv,(char*)0,(char*)0);CHKERRQ(ierr);
  
  if (petsc_involved==1) //use only a cube number of processors
  {
    if (rank==0) std::cout<<"using "<<petscsize<<" of "<<worldsize<<" processes\n"<<std::flush;
    if (rank==0) std::cout<<N<<"^3 unknowns per processor\n"<<std::flush;
    if (rank==0) std::cout<<"total system size: "<<g_N<<"^3\n"<<std::flush;
    if (rank==0) std::cout<<"mesh size: "<<h<<"\n\n"<<std::flush;
  
  
    PETScSolver *petscsolver;
    int g_Ni = g_N; //global number of nodes in each direction
    int g_Nj = g_N;
    int g_Nk = g_N;
    
    int l_Ni[dims[0]];
    int l_Nj[dims[1]];
    int l_Nk[dims[2]];
    for (int i=0; i<procs_per_dim; i++){
      l_Ni[i] = N;  //number of local nodes along the x,y and z axis for each cell,
      l_Nj[i] = N;  //as passed to DMDACreate3d
      l_Nk[i] = N;
    }
    
    int rankpos[3]; //position of rank in process grid; same as PETSc choses for DMDA
    rankpos[0] = rank%dims[0];
    rankpos[1] = (rank/dims[0])%dims[1];
    rankpos[2] = (rank/(dims[0]*dims[1]))%dims[2];
    
    int imin=0,imax,jmin=0,jmax,kmin=0,kmax;
    for (int j=0; j<rankpos[0]; j++)
      imin += l_Ni[j];
    imax = imin+l_Ni[rankpos[0]]-1;
    for (int j=0; j<rankpos[1]; j++)
      jmin += l_Nj[j];
    jmax = jmin+l_Nj[rankpos[1]]-1;
    for (int j = 0; j<rankpos[2]; j++)
      kmin += l_Nk[j];
    kmax = kmin+l_Nk[rankpos[2]]-1;
    
    int Ni = imax-imin+1;
    int Nj = jmax-jmin+1;
    int Nk = kmax-kmin+1;
    int Nall = Ni*Nj*Nk;

    MPI_Barrier(comm);
    petscsolver = new PETScSolver(g_Ni,g_Nj,g_Nk,dims,l_Ni,l_Nj,l_Nk,h);
    
    double *phi,*charge; //electric potential and charge distribution
    int *bound; //used to define Dirichlet BC inside the domain
    phi = new double[Nall];
    for (int i=0; i<Nall; i++)
      phi[i] = 0;
    charge = new double[Nall];
    for (int i=0; i<Nall; i++)
      charge[i] = 1E-18;
    bound = new int[(Ni+2)*(Nj+2)*(Nk+2)]; //needs to include ghost points for matrix assembly
    for (int k=kmin-1; k<=kmax+1; k++) {
      for (int j=jmin-1; j<=jmax+1; j++) {
        for (int i=imin-1; i<=imax+1; i++) {
          if ((i>=0 && i<=1) || (i>=g_Ni-2 && i<=g_Ni-1)
           || (j>=0 && j<=1) || (j>=g_Nj-2 && j<=g_Nj-1)
           || (k>=0 && k<=1) || (k>=g_Nk-2 && k<=g_Nk-1))
            bound[i-imin+1+(j-jmin+1)*(Ni+2)+(k-kmin+1)*(Ni+2)*(Nj+2)] = 1;
          else
            bound[i-imin+1+(j-jmin+1)*(Ni+2)+(k-kmin+1)*(Ni+2)*(Nj+2)] = 0;
        }
      }
    }
    
    MPI_Barrier(comm);
    petscsolver->initialize_matrix(bound);
    
    petscsolver->total_number_of_solve_iterations = iterations;
    petscsolver->initsolve(phi,charge);
    
    for (int noi=1; noi<=iterations; noi++) { //repeated calls to the solve function with the same 
      for (int i=0; i<Nall; i++)              //input to simulate the behavior in a PIC context
        phi[i] = 0;
      petscsolver->solve(noi,phi,charge);
    }
    
    delete[] phi;
    delete[] charge;
    delete[] bound;
    delete petscsolver;
  }

  
  MPI_Barrier(MPI_COMM_WORLD);
  
  PetscFinalize();
  
  MPI_Finalize();
  
  return 0;
}

