#include "petscsolver.h"

PETScSolver::PETScSolver(int g_Ni,int g_Nj,int g_Nk,int* dims,int* l_Ni,int* l_Nj,int* l_Nk,double h) {
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&worldsize);

  ih = 1.0/h;

  int rankpos[3];
  rankpos[0] = rank%dims[0];
  rankpos[1] = (rank/dims[0])%dims[1];
  rankpos[2] = (rank/(dims[0]*dims[1]))%dims[2];

  imin = 0;
  for (int j=0; j<rankpos[0]; j++)
    imin += l_Ni[j];
  imax = imin+l_Ni[rankpos[0]]-1;

  jmin = 0;
  for (int j=0; j<rankpos[1]; j++)
    jmin += l_Nj[j];
  jmax = jmin+l_Nj[rankpos[1]]-1;

  kmin = 0;
  for (int j=0; j<rankpos[2]; j++)
    kmin += l_Nk[j];
  kmax = kmin+l_Nk[rankpos[2]]-1;

  Ni = l_Ni[rankpos[0]];
  Nj = l_Nj[rankpos[1]];
  Nk = l_Nk[rankpos[2]];
  Nall = Ni*Nj*Nk;
  
  initialize_petsc(g_Ni,g_Nj,g_Nk,dims,l_Ni,l_Nj,l_Nk);
  
  bound = new int[(Ni+2)*(Nj+2)*(Nk+2)];
  memset(bound,0,(Ni+2)*(Nj+2)*(Nk+2)*sizeof(int));
}

int PETScSolver::initialize_petsc(int g_Ni,int g_Nj,int g_Nk,int* dims,int* l_Ni,int* l_Nj,int* l_Nk) {
  PetscErrorCode ierr;
  ierr=KSPCreate(PETSC_COMM_WORLD,&ksp);CHKERRQ(ierr);

  ierr=DMDACreate3d(PETSC_COMM_WORLD,DM_BOUNDARY_GHOSTED,DM_BOUNDARY_GHOSTED,DM_BOUNDARY_GHOSTED,
    DMDA_STENCIL_STAR,g_Ni,g_Nj,g_Nk,dims[0],dims[1],dims[2],1,1,l_Ni,l_Nj,l_Nk,&da);CHKERRQ(ierr);
  ierr=DMSetFromOptions(da);CHKERRQ(ierr);
  ierr=DMSetUp(da);CHKERRQ(ierr);
  
  ierr=KSPSetDM(ksp,da);CHKERRQ(ierr);

  ierr=DMCreateGlobalVector(da,&b);CHKERRQ(ierr);
  ierr=VecDuplicate(b,&x);CHKERRQ(ierr);

  ierr=VecSet(x,0);CHKERRQ(ierr);
  ierr=VecSet(b,0);CHKERRQ(ierr);

  return ierr;
}

PETScSolver::~PETScSolver() {
  if (rank==0) std::cout<<"\nTime in solve():      "<<time_in_solve<<" s\n"
                          <<"Time in KSPSolve():   "<<time_in_KSPSolve<<" s ("
                          <<time_in_KSPSolve/time_in_solve*100<<"%)\n\n"<<std::flush;
  if (rank==0) std::cout<<"Number of   KSP iterations (total): "<<total_number_of_KSP_iterations<<"\n"<<std::flush;
  if (rank==0) std::cout<<"Number of solve iterations (total): "<<i_solve<<" (ratio: "<<std::fixed<<std::setprecision(2)
                          <<float(total_number_of_KSP_iterations)/float(i_solve)<<")\n\n"<<std::flush;

  delete[] bound;

  VecDestroy(&x);
  VecDestroy(&b);
  DMDestroy(&da);
  KSPDestroy(&ksp);
}

void PETScSolver::initialize_matrix(int* _bound) {
  for (int i=0; i<(Ni+2)*(Nj+2)*(Nk+2); i++)
    bound[i] = _bound[i];
    
  KSPSetComputeOperators(ksp,ComputeMatrix,(void*)bound);
}

void PETScSolver::solve(int timestep,double* phi,double* charge) {
  double t0 = MPI_Wtime();

  if (i_solve++==0) {
      PetscLogStageRegister("Remaining Solves",&stage[1]);
      PetscLogStagePush(stage[1]);
  }

  PetscScalar *barray;
  VecGetArray(b,&barray);
  for (int k=kmin; k<=kmax; k++)
    for (int j=jmin; j<=jmax; j++)
      for (int i=imin; i<=imax; i++) {
        int idx = i-imin+(j-jmin)*Ni+(k-kmin)*Ni*Nj;
        int idx2 = i-imin+1+(j-jmin+1)*(Ni+2)+(k-kmin+1)*(Ni+2)*(Ni+2);
          if (bound[idx2]==0)
            barray[idx] = charge[idx]*ih*iepsilon0;
          else
            barray[idx] = phi[idx];
      }
  VecRestoreArray(b,&barray);

  KSPSetInitialGuessNonzero(ksp,PETSC_TRUE);

  PetscScalar *xarray;
  VecGetArray(x,&xarray);
  for (int i=0; i<Nall; i++)
    xarray[i] = phi[i];
  VecRestoreArray(x,&xarray);
  
  double t1 = MPI_Wtime();
  KSPSolve(ksp,b,x);
  time_in_KSPSolve += MPI_Wtime()-t1;

  int its;
  KSPGetIterationNumber(ksp,&its);
  total_number_of_KSP_iterations += its;
    
  if (timestep % 100==0 || timestep<=10 || (timestep%10==0 && timestep<100)) {
    if (rank==0) std::cout<<"solve "<<timestep<<": "<<its<<" iterations\n"<<std::flush;
  }
  KSPGetSolution(ksp,&x);
  
  VecGetArray(x,&xarray);
  for (int i=0; i<Nall; i++)
    phi[i] = xarray[i];
  VecRestoreArray(x,&xarray);
  
  if (timestep==total_number_of_solve_iterations)
    PetscLogStagePop();

  time_in_solve += MPI_Wtime()-t0;
}

void PETScSolver::initsolve(double *phi,double *charge) {  
  PetscLogStageRegister("First Solve",&stage[0]);
  PetscLogStagePush(stage[0]);
  
  KSPSetInitialGuessNonzero(ksp,PETSC_TRUE);
  KSPSetFromOptions(ksp);

  PetscScalar *xarray;
  VecGetArray(x,&xarray);
  for (int i=0; i<Nall; i++)
    xarray[i] = phi[i];
  VecRestoreArray(x,&xarray);

  PetscScalar *barray;
  VecGetArray(b,&barray);
  for (int k=kmin; k<=kmax; k++)
    for (int j=jmin; j<=jmax; j++)
      for (int i=imin; i<=imax; i++) {
        int idx = i-imin+(j-jmin)*Ni+(k-kmin)*Ni*Nj;
        int idx2 = i-imin+1+(j-jmin+1)*(Ni+2)+(k-kmin+1)*(Ni+2)*(Ni+2);
          if (bound[idx2]==0)
            barray[idx] = charge[idx] * ih * iepsilon0;
          else
            barray[idx] = phi[idx];
      }
  VecRestoreArray(b,&barray);

  KSPSolve(ksp,b,x);

  int its;
  KSPGetIterationNumber(ksp,&its);
  if (rank==0) std::cout<<"initsolve: "<<its<<" iterations\n"<<std::flush;
  KSPGetSolution(ksp,&x);
  
  VecGetArray(x,&xarray);
  for (int i=0; i<Nall; i++)
    phi[i] = xarray[i];
  VecRestoreArray(x,&xarray);
  
  KSPConvergedReason reason;
  KSPGetConvergedReason(ksp,&reason);
  if (reason<0) {
    if (rank==0) {
      std::cout<<"Initsolve: no convergence\n"<<std::flush;
      if (reason==0) std::cout<<"         Reason: KSP_CONVERGED_ITERATING\n"<<std::flush;
      if (reason==1) std::cout<<"         Reason: KSP_CONVERGED_RTOL_NORMAL\n"<<std::flush;
      if (reason==2) std::cout<<"         Reason: KSP_CONVERGED_RTOL\n"<<std::flush;
      if (reason==3) std::cout<<"         Reason: KSP_CONVERGED_ATOL\n"<<std::flush;
      if (reason==4) std::cout<<"         Reason: KSP_CONVERGED_ITS\n"<<std::flush;
      if (reason==5) std::cout<<"         Reason: KSP_CONVERGED_CG_NEG_CURVE\n"<<std::flush;
      if (reason==6) std::cout<<"         Reason: KSP_CONVERGED_CG_CONSTRAINED\n"<<std::flush;
      if (reason==7) std::cout<<"         Reason: KSP_CONVERGED_STEP_LENGTH\n"<<std::flush;
      if (reason==8) std::cout<<"         Reason: KSP_CONVERGED_HAPPY_BREAKDOWN\n"<<std::flush;
      if (reason==-2) std::cout<<"         Reason: KSP_DIVERGED_NULL\n"<<std::flush;
      if (reason==-3) std::cout<<"         Reason: KSP_DIVERGED_ITS\n"<<std::flush;
      if (reason==-4) std::cout<<"         Reason: KSP_DIVERGED_DTOL\n"<<std::flush;
      if (reason==-5) std::cout<<"         Reason: KSP_DIVERGED_BREAKDOWN\n"<<std::flush;
      if (reason==-6) std::cout<<"         Reason: KSP_DIVERGED_BREAKDOWN_BICG\n"<<std::flush;
      if (reason==-7) std::cout<<"         Reason: KSP_DIVERGED_NONSYMMETRIC\n"<<std::flush;
      if (reason==-8) std::cout<<"         Reason: KSP_DIVERGED_INDEFINITE_PC\n"<<std::flush;
      if (reason==-9) std::cout<<"         Reason: KSP_DIVERGED_NANORINF\n"<<std::flush;
      if (reason==-10) std::cout<<"         Reason: KSP_DIVERGED_INDEFINITE_MAT\n"<<std::flush;
      if (reason==-11) std::cout<<"         Reason: KSP_DIVERGED_PCSETUP_FAILED\n"<<std::flush;
    }
    PetscFinalize();
    MPI_Finalize();
    exit(101);
  }
  
  PetscLogStagePop();
}

PetscErrorCode ComputeMatrix(KSP ksp,Mat A,Mat B,void *ctx) {
  DM             da;
  PetscErrorCode ierr;
  PetscInt       i,j,k,mx,my,mz,xm,ym,zm,xs,ys,zs;
  PetscScalar    v[7];
  MatStencil     row,col[7];

  PetscFunctionBeginUser;
  ierr    = KSPGetDM(ksp,&da);CHKERRQ(ierr);
  ierr    = DMDAGetInfo(da,0,&mx,&my,&mz,0,0,0,0,0,0,0,0,0);CHKERRQ(ierr);
  ierr    = DMDAGetCorners(da,&xs,&ys,&zs,&xm,&ym,&zm);CHKERRQ(ierr);

  int Ni_g = xm+2;
  int Nj_g = ym+2;
  int *bound = (int *)ctx;

  for (k=zs; k<zs+zm; k++) {
    for (j=ys; j<ys+ym; j++) {
      for (i=xs; i<xs+xm; i++) {
        row.i = i; row.j = j; row.k = k;
        if (bound[(i-xs+1)+(j-ys+1)*Ni_g+(k-zs+1)*Ni_g*Nj_g]==1) {
          v[0] = 1.0;
          ierr = MatSetValuesStencil(B,1,&row,1,&row,v,INSERT_VALUES);CHKERRQ(ierr);
        } else if (i==0 || j==0 || k==0 || i==mx-1 || j==my-1 || k==mz-1) {
          v[0] = 1.0;
          ierr = MatSetValuesStencil(B,1,&row,1,&row,v,INSERT_VALUES);CHKERRQ(ierr);
        } else {
          int idx = 0;
          if (bound[(i-xs+1)+(j-ys+1)*Ni_g+(k-zs+1-1)*Ni_g*Nj_g]!=1) {
            v[idx] = -1; col[idx].i = i;     col[idx].j = j;     col[idx].k = k-1;
            idx++;
          }
          if (bound[(i-xs+1)+(j-ys+1-1)*Ni_g+(k-zs+1)*Ni_g*Nj_g]!=1) {
            v[idx] = -1; col[idx].i = i;     col[idx].j = j-1;   col[idx].k = k;
            idx++;
          }
          if (bound[(i-xs+1-1)+(j-ys+1)*Ni_g+(k-zs+1)*Ni_g*Nj_g]!=1) {
            v[idx] = -1; col[idx].i = i-1;   col[idx].j = j;     col[idx].k = k;
            idx++;
          }
          v[idx] =  6;   col[idx].i = row.i; col[idx].j = row.j; col[idx].k = row.k;
          idx++;
          if (bound[(i-xs+1+1)+(j-ys+1)*Ni_g+(k-zs+1)*Ni_g*Nj_g]!=1) {
            v[idx] = -1; col[idx].i = i+1;   col[idx].j = j;   col[idx].k = k;
            idx++;
          }
          if (bound[(i-xs+1)+(j-ys+1+1)*Ni_g+(k-zs+1)*Ni_g*Nj_g]!=1) {
            v[idx] = -1; col[idx].i = i;     col[idx].j = j+1;   col[idx].k = k;
            idx++;
          }
          if (bound[(i-xs+1)+(j-ys+1)*Ni_g+(k-zs+1+1)*Ni_g*Nj_g]!=1) {
            v[idx] = -1; col[idx].i = i;     col[idx].j = j;     col[idx].k = k+1;
            idx++;
          }
          ierr = MatSetValuesStencil(B,1,&row,idx,col,v,INSERT_VALUES);CHKERRQ(ierr);
        }
      }
    }
  }

  ierr   = MatAssemblyBegin(B,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr   = MatAssemblyEnd(B,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

